import os
from unittest import mock
from unittest.mock import *

import pytest

from rok4.exceptions import *
from rok4.storage import disconnect_ceph_clients
from rok4.vector import *


@mock.patch.dict(os.environ, {}, clear=True)
def test_missing_env():
    disconnect_ceph_clients()
    with pytest.raises(MissingEnvironmentError):
        Vector.from_file("ceph:///ign_std/vector.shp")


@mock.patch("rok4.vector.copy", side_effect=StorageError("CEPH", "Not found"))
def test_wrong_file(mocked_copy):
    with pytest.raises(StorageError):
        Vector.from_file("ceph:///vector.geojson")


def test_wrong_format():
    with pytest.raises(Exception) as exc:
        Vector.from_file("ceph:///vector.tif")
    assert str(exc.value) == "This format of file cannot be loaded"


@mock.patch("rok4.vector.ogr.Open", return_value="not a shape")
def test_wrong_content(mocked_copy):
    with pytest.raises(Exception) as exc:
        Vector.from_file("file:///vector.shp")
    assert str(exc.value) == "The content of file:///vector.shp cannot be read"


@mock.patch("rok4.vector.copy")
@mock.patch("rok4.vector.ogr.Open", return_value="not a shape")
def test_wrong_content_ceph(mocked_open, mocked_copy):
    with pytest.raises(Exception) as exc:
        Vector.from_file("file:///vector.shp")
    assert str(exc.value) == "The content of file:///vector.shp cannot be read"


def test_ok_csv1():
    try:
        vector_csv1 = Vector.from_file(
            "file://tests/fixtures/vector.csv",
            csv={"delimiter": ";", "column_x": "x", "column_y": "y"},
        )
        assert (
            str(vector_csv1.layers)
            == "[('vector', 3, [('id', 'String'), ('x', 'String'), ('y', 'String')])]"
        )
    except Exception as exc:
        assert False, f"Vector creation raises an exception: {exc}"


def test_ok_csv2():
    try:
        vector_csv2 = Vector.from_file(
            "file://tests/fixtures/vector2.csv", csv={"delimiter": ";", "column_wkt": "WKT"}
        )
        assert str(vector_csv2.layers) == "[('vector2', 1, [('id', 'String'), ('WKT', 'String')])]"
    except Exception as exc:
        assert False, f"Vector creation raises an exception: {exc}"


def test_ok_geojson():
    try:
        vector = Vector.from_file("file://tests/fixtures/vector.geojson")
        assert (
            str(vector.layers)
            == "[('vector', 1, [('id', 'String'), ('id_fantoir', 'String'), ('numero', 'Integer'), ('rep', 'String'), ('nom_voie', 'String'), ('code_postal', 'Integer'), ('code_insee', 'Integer'), ('nom_commune', 'String'), ('code_insee_ancienne_commune', 'String'), ('nom_ancienne_commune', 'String'), ('x', 'Real'), ('y', 'Real'), ('lon', 'Real'), ('lat', 'Real'), ('type_position', 'String'), ('alias', 'String'), ('nom_ld', 'String'), ('libelle_acheminement', 'String'), ('nom_afnor', 'String'), ('source_position', 'String'), ('source_nom_voie', 'String'), ('certification_commune', 'Integer'), ('cad_parcelles', 'String')])]"
        )
    except Exception as exc:
        assert False, f"Vector creation raises an exception: {exc}"


def test_ok_gpkg():
    try:
        vector = Vector.from_file("file://tests/fixtures/vector.gpkg")
        assert (
            str(vector.layers)
            == "[('Table1', 2, [('id', 'String')]), ('Table2', 2, [('id', 'Integer'), ('nom', 'String')])]"
        )
    except Exception as exc:
        assert False, f"Vector creation raises an exception: {exc}"


def test_ok_shp():
    try:
        vector = Vector.from_file("file://tests/fixtures/ARRONDISSEMENT.shp")
        assert (
            str(vector.layers)
            == "[('ARRONDISSEMENT', 14, [('ID', 'String'), ('NOM', 'String'), ('INSEE_ARR', 'String'), ('INSEE_DEP', 'String'), ('INSEE_REG', 'String'), ('ID_AUT_ADM', 'String'), ('DATE_CREAT', 'String'), ('DATE_MAJ', 'String'), ('DATE_APP', 'Date'), ('DATE_CONF', 'Date')])]"
        )
    except Exception as exc:
        assert False, f"Vector creation raises an exception: {exc}"


def test_ok_parameters():
    try:
        vector = Vector.from_parameters(
            "file://tests/fixtures/ARRONDISSEMENT.shp",
            (1, 2, 3, 4),
            [
                (
                    "ARRONDISSEMENT",
                    14,
                    [
                        ("ID", "String"),
                        ("NOM", "String"),
                        ("INSEE_ARR", "String"),
                        ("INSEE_DEP", "String"),
                        ("INSEE_REG", "String"),
                        ("ID_AUT_ADM", "String"),
                        ("DATE_CREAT", "String"),
                        ("DATE_MAJ", "String"),
                        ("DATE_APP", "Date"),
                        ("DATE_CONF", "Date"),
                    ],
                )
            ],
        )
        assert str(vector.path) == "file://tests/fixtures/ARRONDISSEMENT.shp"
    except Exception as exc:
        assert False, f"Vector creation raises an exception: {exc}"



class MissingAttributeError(Exception):
    """
    Exception raised when an attribute is missing in a file
    """

    def __init__(self, path, missing):
        self.path = path
        self.missing = missing
        super().__init__(f"Missing attribute {missing} in '{path}'")


class MissingEnvironmentError(Exception):
    """
    Exception raised when a needed environment variable is not defined
    """

    def __init__(self, missing):
        self.missing = missing
        super().__init__(f"Missing environment variable {missing}")


class StorageError(Exception):
    """
    Exception raised when an issue occured when using a storage
    """

    def __init__(self, type, issue):
        self.type = type
        self.issue = issue
        super().__init__(f"Issue occured using a {type} storage : {issue}")


class FormatError(Exception):
    """
    Exception raised when a format is expected but not respected
    """

    def __init__(self, expected_format, content, issue):
        self.expected_format = expected_format
        self.content = content
        self.issue = issue
        super().__init__(f"Expected format {expected_format} to read '{content}' : {issue}")



#! python3  # noqa: E265

# standard lib
from enum import Enum


class PyramidType(Enum):
    """Pyramid's data type"""

    RASTER = "RASTER"
    VECTOR = "VECTOR"


class SlabType(Enum):
    """Slab's type"""

    DATA = "DATA"  # Slab of data, raster or vector
    MASK = "MASK"  # Slab of mask, only for raster pyramid, image with one band : 0 is nodata, other values are data


class StorageType(Enum):
    """Storage type and path's protocol"""

    CEPH = "ceph://"
    FILE = "file://"
    HTTP = "http://"
    HTTPS = "https://"
    S3 = "s3://"


class ColorFormat(Enum):
    """A color format enumeration.
    Except from "BIT", the member's name matches
      a common variable format name. The member's value is
      the allocated bit size associated to this format.
    """

    BIT = 1
    UINT8 = 8
    FLOAT32 = 32



import os
from unittest import mock
from unittest.mock import *

from rok4.enums import PyramidType
from rok4.exceptions import *
from rok4.layer import Layer


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.layer.get_data_str",
    return_value='{"pyramids" : [{"bottom_level" : "10","top_level" : "10","path" : "s3://pyramids/SCAN1000.json"}],"title" : "SCAN1000","bbox":{"east": 11.250000000000997,"west": -5.624999999999043,"north": 52.48278022207774,"south": 40.9798980696195},"styles" : ["normal","hypso"],"abstract" : "Diffusion de la donnée BDORTHO","resampling" : "linear","keywords" : ["PM","TIFF_JPG_UINT8"]}',
)
@mock.patch("rok4.layer.Pyramid.from_descriptor")
@mock.patch("rok4.layer.put_data_str", return_value=None)
def test_descriptor_ok(mocked_put_data_str, mocked_pyramid_class, mocked_get_data_str):
    tms_instance = MagicMock()
    tms_instance.srs = "EPSG:3857"

    level_instance = MagicMock()
    level_instance.id = 10
    level_instance.resolution = 1

    pyramid_instance = MagicMock()
    pyramid_instance.raster_specifications = {
        "channels": 3,
        "nodata": "255,255,255",
        "photometric": "rgb",
        "interpolation": "bicubic",
    }
    pyramid_instance.format = "TIFF_JPG_UINT8"
    pyramid_instance.tms = tms_instance
    pyramid_instance.descriptor = "s3://pyramids/SCAN1000.json"
    pyramid_instance.get_levels.return_value = [level_instance]
    mocked_pyramid_class.return_value = pyramid_instance

    try:
        layer = Layer.from_descriptor("s3://layers/SCAN1000.json")
        assert layer.type == PyramidType.RASTER
        mocked_get_data_str.assert_called_once_with("s3://layers/SCAN1000.json")

        layer.write_descriptor("s3://layers_backup/")
        mocked_put_data_str.assert_called_once_with(
            '{"title": "SCAN1000", "abstract": "Diffusion de la donn\\u00e9e BDORTHO", "keywords": ["PM", "TIFF_JPG_UINT8"], "wmts": {"authorized": true}, "tms": {"authorized": true}, "bbox": {"south": 40.9798980696195, "west": -5.624999999999043, "north": 52.48278022207774, "east": 11.250000000000997}, "pyramids": [{"bottom_level": "10", "top_level": "10", "path": "s3://pyramids/SCAN1000.json"}], "wms": {"authorized": true, "crs": ["CRS:84", "IGNF:WGS84G", "EPSG:3857", "EPSG:4258", "EPSG:4326"]}, "styles": ["normal", "hypso"], "resampling": "linear"}',
            "s3://layers_backup/SCAN1000.json",
        )
    except Exception as exc:
        assert False, f"Layer creation from descriptor raises an exception: {exc}"


@mock.patch("rok4.layer.Pyramid.from_descriptor")
@mock.patch("rok4.layer.reproject_bbox", return_value=(0, 0, 100, 100))
@mock.patch("rok4.layer.put_data_str", return_value=None)
def test_parameters_vector_ok(
    mocked_put_data_str, mocked_utils_reproject_bbox, mocked_pyramid_class
):
    tms_instance = MagicMock()
    tms_instance.srs = "EPSG:3857"

    level_instance = MagicMock()
    level_instance.id = 10
    level_instance.resolution = 1
    level_instance.bbox = (0, 0, 100000, 100000)

    pyramid_instance = MagicMock()
    pyramid_instance.format = "TIFF_PBF_MVT"
    pyramid_instance.tms = tms_instance
    pyramid_instance.descriptor = "file:///home/ign/pyramids/SCAN1000.json"
    pyramid_instance.get_levels.return_value = [level_instance]
    mocked_pyramid_class.return_value = pyramid_instance

    try:
        layer = Layer.from_parameters(
            [
                {
                    "path": "file:///home/ign/pyramids/SCAN1000.json",
                    "bottom_level": "10",
                    "top_level": "10",
                }
            ],
            "layername",
            title="title",
            abstract="abstract",
        )
        assert layer.type == PyramidType.VECTOR
        assert layer.geobbox == (0, 0, 100, 100)
        layer.write_descriptor("file:///home/ign/layers/")
        mocked_put_data_str.assert_called_once_with(
            '{"title": "title", "abstract": "abstract", "keywords": ["VECTOR", "layername"], "wmts": {"authorized": true}, "tms": {"authorized": true}, "bbox": {"south": 0, "west": 0, "north": 100, "east": 100}, "pyramids": [{"bottom_level": "10", "top_level": "10", "path": "file:///home/ign/pyramids/SCAN1000.json"}]}',
            "file:///home/ign/layers/layername.json",
        )

    except Exception as exc:
        assert False, f"Layer creation from parameters raises an exception: {exc}"


@mock.patch("rok4.layer.Pyramid.from_descriptor")
@mock.patch("rok4.layer.reproject_bbox", return_value=(0, 0, 100, 100))
@mock.patch("rok4.layer.put_data_str", return_value=None)
def test_parameters_raster_ok(
    mocked_put_data_str, mocked_utils_reproject_bbox, mocked_pyramid_class
):
    tms_instance = MagicMock()
    tms_instance.srs = "EPSG:3857"

    level_instance = MagicMock()
    level_instance.id = 10
    level_instance.resolution = 1
    level_instance.bbox = (0, 0, 100000, 100000)

    pyramid_instance = MagicMock()
    pyramid_instance.format = "TIFF_ZIP_FLOAT32"
    pyramid_instance.raster_specifications = {
        "channels": 1,
        "nodata": "-99999",
        "photometric": "gray",
        "interpolation": "nn",
    }
    pyramid_instance.tms = tms_instance
    pyramid_instance.bottom_level.id = "10"
    pyramid_instance.top_level.id = "10"
    pyramid_instance.descriptor = "file:///home/ign/pyramids/RGEALTI.json"
    pyramid_instance.get_levels.return_value = [level_instance]
    mocked_pyramid_class.return_value = pyramid_instance

    try:
        layer = Layer.from_parameters(
            [{"path": "file:///home/ign/pyramids/RGEALTI.json"}],
            "layername",
            title="title",
            abstract="abstract",
        )
        assert layer.type == PyramidType.RASTER
        assert layer.geobbox == (0, 0, 100, 100)
        layer.write_descriptor("file:///home/ign/layers/")
        mocked_put_data_str.assert_called_once_with(
            '{"title": "title", "abstract": "abstract", "keywords": ["RASTER", "layername"], "wmts": {"authorized": true}, "tms": {"authorized": true}, "bbox": {"south": 0, "west": 0, "north": 100, "east": 100}, "pyramids": [{"bottom_level": "10", "top_level": "10", "path": "file:///home/ign/pyramids/RGEALTI.json"}], "wms": {"authorized": true, "crs": ["CRS:84", "IGNF:WGS84G", "EPSG:3857", "EPSG:4258", "EPSG:4326"]}, "styles": ["normal"], "resampling": "nn"}',
            "file:///home/ign/layers/layername.json",
        )

    except Exception as exc:
        assert False, f"Layer creation from parameters raises an exception: {exc}"



"""Provide functions to read information on raster data

The module contains the following class :

    - Raster - Structure describing raster data.
    - RasterSet - Structure describing a set of raster data.
"""

# -- IMPORTS --

# standard library
import copy
import json
import re
from typing import Dict, Tuple

# 3rd party
from osgeo import gdal, ogr

# package
from rok4.enums import ColorFormat
from rok4.storage import exists, get_osgeo_path, put_data_str
from rok4.utils import compute_bbox, compute_format

# -- GLOBALS --

# Enable GDAL/OGR exceptions
ogr.UseExceptions()
gdal.UseExceptions()


class Raster:
    """A structure describing raster data

    Attributes :
        path (str): path to the file/object (ex:
          file:///path/to/image.tif or s3://bucket/path/to/image.tif)
        bbox (Tuple[float, float, float, float]): bounding rectange
          in the data projection
        bands (int): number of color bands (or channels)
        format (ColorFormat): numeric variable format for color values.
          Bit depth, as bits per channel, can be derived from it.
        mask (str): path to the associated mask file or object, if any,
          or None (same path as the image, but with a ".msk" extension
          and TIFF format. ex:
          file:///path/to/image.msk or s3://bucket/path/to/image.msk)
        dimensions (Tuple[int, int]): image width and height, in pixels
    """

    def __init__(self) -> None:
        self.bands = None
        self.bbox = (None, None, None, None)
        self.dimensions = (None, None)
        self.format = None
        self.mask = None
        self.path = None

    @classmethod
    def from_file(cls, path: str) -> "Raster":
        """Creates a Raster object from an image

        Args:
            path (str): path to the image file/object

        Examples:

            Loading informations from a file stored raster TIFF image

                from rok4.raster import Raster

                try:
                    raster = Raster.from_file(
                        "file:///data/SC1000/0040_6150_L93.tif"
                    )

                except Exception as e:
                    print(f"Cannot load information from image : {e}")

        Raises:
            RuntimeError: raised by OGR/GDAL if anything goes wrong
            NotImplementedError: Storage type not handled

        Returns:
            Raster: a Raster instance
        """
        if not exists(path):
            raise Exception(f"No file or object found at path '{path}'.")

        self = cls()

        work_image_path = get_osgeo_path(path)

        image_datasource = gdal.Open(work_image_path)
        self.path = path

        path_pattern = re.compile("(/[^/]+?)[.][a-zA-Z0-9_-]+$")
        mask_path = path_pattern.sub("\\1.msk", path)

        if exists(mask_path):
            work_mask_path = get_osgeo_path(mask_path)
            mask_driver = gdal.IdentifyDriver(work_mask_path).ShortName
            if "GTiff" != mask_driver:
                message = (
                    f"Mask file '{mask_path}' is not a TIFF image."
                    + f" (GDAL driver : '{mask_driver}'"
                )
                raise Exception(message)
            self.mask = mask_path
        else:
            self.mask = None

        self.bbox = compute_bbox(image_datasource)
        self.bands = image_datasource.RasterCount
        self.format = compute_format(image_datasource, path)
        self.dimensions = (image_datasource.RasterXSize, image_datasource.RasterYSize)

        return self

    @classmethod
    def from_parameters(
        cls,
        path: str,
        bands: int,
        bbox: Tuple[float, float, float, float],
        dimensions: Tuple[int, int],
        format: ColorFormat,
        mask: str = None,
    ) -> "Raster":
        """Creates a Raster object from parameters

        Args:
            path (str): path to the file/object (ex:
              file:///path/to/image.tif or s3://bucket/image.tif)
            bands (int): number of color bands (or channels)
            bbox (Tuple[float, float, float, float]): bounding rectange
              in the data projection
            dimensions (Tuple[int, int]): image width and height
              expressed in pixels
            format (ColorFormat): numeric format for color values.
              Bit depth, as bits per channel, can be derived from it.
            mask (str, optionnal): path to the associated mask, if any,
              or None (same path as the image, but with a
              ".msk" extension and TIFF format. ex:
              file:///path/to/image.msk or s3://bucket/image.msk)

        Examples:

            Loading informations from parameters, related to
              a TIFF main image coupled to a TIFF mask image

                from rok4.raster import Raster

                try:
                    raster = Raster.from_parameters(
                      path="file:///data/SC1000/_0040_6150_L93.tif",
                      mask="file:///data/SC1000/0040_6150_L93.msk",
                      bands=3,
                      format=ColorFormat.UINT8,
                      dimensions=(2000, 2000),
                      bbox=(40000.000, 5950000.000,
                            240000.000, 6150000.000)
                    )

                except Exception as e:
                    print(
                      f"Cannot load information from parameters : {e}"
                    )

        Raises:
            KeyError: a mandatory argument is missing

        Returns:
            Raster: a Raster instance
        """
        self = cls()

        self.path = path
        self.bands = bands
        self.bbox = bbox
        self.dimensions = dimensions
        self.format = format
        self.mask = mask
        return self


class RasterSet:
    """A structure describing a set of raster data

    Attributes :
        raster_list (List[Raster]): List of Raster instances in the set
        colors (List[Dict]): List of color properties for each raster
              instance. Contains only one element if
              the set is homogenous.
            Element properties:
                bands (int): number of color bands (or channels)
                format (ColorFormat): numeric variable format for
                  color values. Bit depth, as bits per channel,
                  can be derived from it.
        srs (str): Name of the set's spatial reference system
        bbox (Tuple[float, float, float, float]): bounding rectange
          in the data projection, enclosing the whole set
    """

    def __init__(self) -> None:
        self.bbox = (None, None, None, None)
        self.colors = []
        self.raster_list = []
        self.srs = None

    @classmethod
    def from_list(cls, path: str, srs: str) -> "RasterSet":
        """Instanciate a RasterSet from an images list path and a srs

        Args:
            path (str): path to the images list file or object
              (each line in this list contains the path to
              an image file or object in the set)

        Examples:

            Loading informations from a file stored list

                from rok4.raster import RasterSet

                try:
                    raster_set = RasterSet.from_list(
                      path="file:///data/SC1000.list",
                      srs="EPSG:3857"
                    )

                except Exception as e:
                    print(
                      f"Cannot load information from list file : {e}"
                    )

        Raises:
            RuntimeError: raised by OGR/GDAL if anything goes wrong
            NotImplementedError: Storage type not handled

        Returns:
            RasterSet: a RasterSet instance
        """
        self = cls()
        self.srs = srs

        local_list_path = get_osgeo_path(path)
        image_list = []
        with open(file=local_list_path) as list_file:
            for line in list_file:
                image_path = line.strip(" \t\n\r")
                image_list.append(image_path)

        temp_bbox = [None, None, None, None]
        for image_path in image_list:
            raster = Raster.from_file(image_path)
            self.raster_list.append(raster)
            if temp_bbox == [None, None, None, None]:
                for i in range(0, 4, 1):
                    temp_bbox[i] = raster.bbox[i]
            else:
                if temp_bbox[0] > raster.bbox[0]:
                    temp_bbox[0] = raster.bbox[0]
                if temp_bbox[1] > raster.bbox[1]:
                    temp_bbox[1] = raster.bbox[1]
                if temp_bbox[2] < raster.bbox[2]:
                    temp_bbox[2] = raster.bbox[2]
                if temp_bbox[3] < raster.bbox[3]:
                    temp_bbox[3] = raster.bbox[3]
            color_dict = {"bands": raster.bands, "format": raster.format}
            if color_dict not in self.colors:
                self.colors.append(color_dict)
        self.bbox = tuple(temp_bbox)
        return self

    @classmethod
    def from_descriptor(cls, path: str) -> "RasterSet":
        """Creates a RasterSet object from a descriptor file or object

        Args:
            path (str): path to the descriptor file or object

        Examples:

            Loading informations from a file stored descriptor

                from rok4.raster import RasterSet

                try:
                    raster_set = RasterSet.from_descriptor(
                      "file:///data/images/descriptor.json"
                    )

                except Exception as e:
                    message = ("Cannot load information from "
                              + f"descriptor file : {e}")
                    print(message)

        Raises:
            RuntimeError: raised by OGR/GDAL if anything goes wrong
            NotImplementedError: Storage type not handled

        Returns:
            RasterSet: a RasterSet instance
        """
        self = cls()
        descriptor_path = get_osgeo_path(path)
        with open(file=descriptor_path) as file_handle:
            raw_content = file_handle.read()
        serialization = json.loads(raw_content)
        self.srs = serialization["srs"]
        self.raster_list = []
        for raster_dict in serialization["raster_list"]:
            parameters = copy.deepcopy(raster_dict)
            parameters["bbox"] = tuple(raster_dict["bbox"])
            parameters["dimensions"] = tuple(raster_dict["dimensions"])
            parameters["format"] = ColorFormat[raster_dict["format"]]
            self.raster_list.append(Raster.from_parameters(**parameters))
        self.bbox = tuple(serialization["bbox"])
        self.colors = []
        for color_dict in serialization["colors"]:
            color_item = copy.deepcopy(color_dict)
            color_item["format"] = ColorFormat[color_dict["format"]]
            self.colors.append(color_item)
        return self

    @property
    def serializable(self) -> Dict:
        """Get the dict version of the raster set, descriptor compliant

        Returns:
            Dict: descriptor structured object description
        """
        serialization = {"bbox": list(self.bbox), "srs": self.srs, "colors": [], "raster_list": []}
        for color in self.colors:
            color_serial = {"bands": color["bands"], "format": color["format"].name}
            serialization["colors"].append(color_serial)
        for raster in self.raster_list:
            raster_dict = {
                "path": raster.path,
                "dimensions": list(raster.dimensions),
                "bbox": list(raster.bbox),
                "bands": raster.bands,
                "format": raster.format.name,
            }
            if raster.mask is not None:
                raster_dict["mask"] = raster.mask
            serialization["raster_list"].append(raster_dict)
        return serialization

    def write_descriptor(self, path: str = None) -> None:
        """Print raster set's descriptor as JSON

        Args:
            path (str, optional): Complete path (file or object)
              where to print the raster set's JSON. Defaults to None,
              JSON is printed to standard output.
        """
        content = json.dumps(self.serializable, sort_keys=True)
        if path is None:
            print(content)
        else:
            put_data_str(content, path)



"""Provide functions to use read or write

Available storage types are :
- S3 (path are preffixed with `s3://`)
- CEPH (path are prefixed with `ceph://`)
- FILE (path are prefixed with `file://`, but it is the default paths' interpretation)
- HTTP (path are prefixed with `http://`)
- HTTPS (path are prefixed with `https://`)

According to functions, all storage types are not necessarily available.

Using CEPH storage requires environment variables :
- ROK4_CEPH_CONFFILE
- ROK4_CEPH_USERNAME
- ROK4_CEPH_CLUSTERNAME

Using S3 storage requires environment variables :
- ROK4_S3_KEY
- ROK4_S3_SECRETKEY
- ROK4_S3_URL

To use several S3 clusters, each environment variable have to contain a list (comma-separated), with the same number of elements

Example: work with 2 S3 clusters:

- ROK4_S3_KEY=KEY1,KEY2
- ROK4_S3_SECRETKEY=SKEY1,SKEY2
- ROK4_S3_URL=https://s3.storage.fr,https://s4.storage.fr

To precise the cluster to use, bucket name should be bucket_name@s3.storage.fr or bucket_name@s4.storage.fr. If no host is defined (no @) in the bucket name, first S3 cluster is used
"""

import hashlib
import os
import re
import tempfile
from shutil import copyfile
from typing import Dict, Tuple, Union

import boto3
import botocore.exceptions
import requests
from osgeo import gdal

# conditional import
try:
    import rados

    CEPH_RADOS_AVAILABLE: bool = True
except ImportError:
    CEPH_RADOS_AVAILABLE: bool = False
    rados = None

# package
from rok4.enums import StorageType
from rok4.exceptions import MissingEnvironmentError, StorageError

# -- GLOBALS --

# Enable GDAL/OGR exceptions
gdal.UseExceptions()

__CEPH_CLIENT = None
__CEPH_IOCTXS = {}
__OBJECT_SYMLINK_SIGNATURE = "SYMLINK#"
__S3_CLIENTS = {}
__S3_DEFAULT_CLIENT = None


def __get_s3_client(bucket_name: str) -> Tuple[Dict[str, Union["boto3.client", str]], str, str]:
    """Get the S3 client

    Create it if not already done

    Args:
        bucket_name (str): S3 bucket name. Could be just the bucket name, or <bucket name>@<cluster host>

    Raises:
        MissingEnvironmentError: Missing S3 storage informations
        StorageError: S3 client configuration issue

    Returns:
        Tuple[Dict[str, Union['boto3.client',str]], str, str]: the S3 informations (client, host, key, secret) and the simple bucket name
    """

    global __S3_CLIENTS, __S3_DEFAULT_CLIENT

    if not __S3_CLIENTS:
        # C'est la première fois qu'on cherche à utiliser le stockage S3, chargeons les informations depuis les variables d'environnement
        try:
            keys = os.environ["ROK4_S3_KEY"].split(",")
            secret_keys = os.environ["ROK4_S3_SECRETKEY"].split(",")
            urls = os.environ["ROK4_S3_URL"].split(",")

            if len(keys) != len(secret_keys) or len(keys) != len(urls):
                raise StorageError(
                    "S3",
                    "S3 informations in environment variables are inconsistent : same number of element in each list is required",
                )

            for i in range(len(keys)):
                h = re.sub("https?://", "", urls[i])

                if h in __S3_CLIENTS:
                    raise StorageError("S3", "A S3 cluster is defined twice (based on URL)")

                __S3_CLIENTS[h] = {
                    "client": boto3.client(
                        "s3",
                        aws_access_key_id=keys[i],
                        aws_secret_access_key=secret_keys[i],
                        endpoint_url=urls[i],
                    ),
                    "key": keys[i],
                    "secret_key": secret_keys[i],
                    "url": urls[i],
                    "host": h,
                }

                if i == 0:
                    # Le premier cluster est celui par défaut
                    __S3_DEFAULT_CLIENT = h

        except KeyError as e:
            raise MissingEnvironmentError(e)
        except Exception as e:
            raise StorageError("S3", e)

    try:
        host = bucket_name.split("@")[1]
    except IndexError:
        host = __S3_DEFAULT_CLIENT

    bucket_name = bucket_name.split("@")[0]

    if host not in __S3_CLIENTS:
        raise StorageError("S3", f"Unknown S3 cluster, according to host '{host}'")

    return __S3_CLIENTS[host], bucket_name


def disconnect_s3_clients() -> None:
    """Clean S3 clients"""

    global __S3_CLIENTS, __S3_DEFAULT_CLIENT
    __S3_CLIENTS = {}
    __S3_DEFAULT_CLIENT = None


def __get_ceph_ioctx(pool: str) -> "rados.Ioctx":
    """Get the CEPH IO context

    Create it (client and context) if not already done

    Args:
        pool (str): CEPH pool's name

    Raises:
        MissingEnvironmentError: Missing CEPH storage informations
        StorageError: CEPH IO context configuration issue

    Returns:
        rados.Ioctx: IO ceph context
    """
    global __CEPH_CLIENT, __CEPH_IOCTXS

    if __CEPH_CLIENT is None:
        try:
            __CEPH_CLIENT = rados.Rados(
                conffile=os.environ["ROK4_CEPH_CONFFILE"],
                clustername=os.environ["ROK4_CEPH_CLUSTERNAME"],
                name=os.environ["ROK4_CEPH_USERNAME"],
            )

            __CEPH_CLIENT.connect()

        except KeyError as e:
            raise MissingEnvironmentError(e)
        except Exception as e:
            raise StorageError("CEPH", e)

    if pool not in __CEPH_IOCTXS:
        try:
            __CEPH_IOCTXS[pool] = __CEPH_CLIENT.open_ioctx(pool)
        except Exception as e:
            raise StorageError("CEPH", e)

    return __CEPH_IOCTXS[pool]


def disconnect_ceph_clients() -> None:
    """Clean CEPH clients"""
    global __CEPH_CLIENT, __CEPH_IOCTXS
    __CEPH_CLIENT = None
    __CEPH_IOCTXS = {}


def get_infos_from_path(path: str) -> Tuple[StorageType, str, str, str]:
    """Extract storage type, the unprefixed path, the container and the basename from path (Default: FILE storage)

    For a FILE storage, the tray is the directory and the basename is the file name.

    For an object storage (CEPH or S3), the tray is the bucket or the pool and the basename is the object name.
    For a S3 bucket, format can be <bucket name>@<cluster name> to use several clusters. Cluster name is the host (without protocol)

    Args:
        path (str): path to analyse

    Returns:
        Tuple[StorageType, str, str, str]: storage type, unprefixed path, the container and the basename
    """

    if path.startswith("s3://"):
        bucket_name, object_name = path[5:].split("/", 1)
        return StorageType.S3, path[5:], bucket_name, object_name
    elif path.startswith("ceph://"):
        pool_name, object_name = path[7:].split("/", 1)
        return StorageType.CEPH, path[7:], pool_name, object_name
    elif path.startswith("file://"):
        return StorageType.FILE, path[7:], os.path.dirname(path[7:]), os.path.basename(path[7:])
    elif path.startswith("http://"):
        return StorageType.HTTP, path[7:], os.path.dirname(path[7:]), os.path.basename(path[7:])
    elif path.startswith("https://"):
        return StorageType.HTTPS, path[8:], os.path.dirname(path[8:]), os.path.basename(path[8:])
    else:
        return StorageType.FILE, path, os.path.dirname(path), os.path.basename(path)


def get_path_from_infos(storage_type: StorageType, *args) -> str:
    """Write full path from elements

    Prefixed wih storage's type, elements are joined with a slash

    Args:
        storage_type (StorageType): Storage's type for path

    Returns:
        str: Full path
    """
    return f"{storage_type.value}{os.path.join(*args)}"


def hash_file(path: str) -> str:
    """Process MD5 sum of the provided file

    Args:
        path (str): path to file

    Returns:
        str: hexadeimal MD5 sum
    """

    checker = hashlib.md5()

    with open(path, "rb") as file:
        chunk = 0
        while chunk != b"":
            chunk = file.read(65536)
            checker.update(chunk)

    return checker.hexdigest()


def get_data_str(path: str) -> str:
    """Load full data into a string

    Args:
        path (str): path to data

    Raises:
        MissingEnvironmentError: Missing object storage informations
        StorageError: Storage read issue
        FileNotFoundError: File or object does not exist

    Returns:
        str: Data content
    """

    return get_data_binary(path).decode("utf-8")


def get_data_binary(path: str, range: Tuple[int, int] = None) -> str:
    """Load data into a binary string

    Args:
        path (str): path to data
        range (Tuple[int, int], optional): offset and size, to make a partial read. Defaults to None.

    Raises:
        MissingEnvironmentError: Missing object storage informations
        StorageError: Storage read issue
        FileNotFoundError: File or object does not exist

    Returns:
        str: Data binary content
    """
    storage_type, path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        try:
            if range is None:
                data = (
                    s3_client["client"]
                    .get_object(
                        Bucket=bucket_name,
                        Key=base_name,
                    )["Body"]
                    .read()
                )
            else:
                data = (
                    s3_client["client"]
                    .get_object(
                        Bucket=bucket_name,
                        Key=base_name,
                        Range=f"bytes={range[0]}-{range[0] + range[1] - 1}",
                    )["Body"]
                    .read()
                )

        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                raise FileNotFoundError(f"{storage_type.value}{path}")
            else:
                raise StorageError("S3", e)

        except Exception as e:
            raise StorageError("S3", e)

    elif storage_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(tray_name)

        try:
            if range is None:
                size, mtime = ioctx.stat(base_name)
                data = ioctx.read(base_name, size)
            else:
                data = ioctx.read(base_name, range[1], range[0])

        except rados.ObjectNotFound:
            raise FileNotFoundError(f"{storage_type.value}{path}")

        except Exception as e:
            raise StorageError("CEPH", e)

    elif storage_type == StorageType.FILE:
        try:
            f = open(path, "rb")
            if range is None:
                data = f.read()
            else:
                f.seek(range[0])
                data = f.read(range[1])

            f.close()

        except FileNotFoundError:
            raise FileNotFoundError(f"{storage_type.value}{path}")

        except Exception as e:
            raise StorageError("FILE", e)

    elif storage_type == StorageType.HTTP or storage_type == StorageType.HTTPS:
        if range is None:
            try:
                reponse = requests.get(f"{storage_type.value}{path}", stream=True)
                data = reponse.content
                if reponse.status_code == 404:
                    raise FileNotFoundError(f"{storage_type.value}{path}")
            except Exception as e:
                raise StorageError(storage_type.name, e)
        else:
            raise NotImplementedError

    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to read binary data")

    return data


def put_data_str(data: str, path: str) -> None:
    """Store string data into a file or an object

    UTF-8 encoding is used for bytes conversion

    Args:
        data (str): data to write
        path (str): destination path, where to write data

    Raises:
        MissingEnvironmentError: Missing object storage informations
        StorageError: Storage write issue
    """

    storage_type, path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        try:
            s3_client["client"].put_object(
                Body=data.encode("utf-8"), Bucket=bucket_name, Key=base_name
            )
        except Exception as e:
            raise StorageError("S3", e)

    elif storage_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(tray_name)

        try:
            ioctx.write_full(base_name, data.encode("utf-8"))
        except Exception as e:
            raise StorageError("CEPH", e)

    elif storage_type == StorageType.FILE:
        try:
            f = open(path, "w")
            f.write(data)
            f.close()
        except Exception as e:
            raise StorageError("FILE", e)

    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to write string data")


def get_size(path: str) -> int:
    """Get size of file or object

    Args:
        path (str): path of file/object whom size is asked

    Raises:
        MissingEnvironmentError: Missing object storage informations
        StorageError: Storage read issue

    Returns:
        int: file/object size, in bytes
    """

    storage_type, path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        try:
            size = s3_client["client"].head_object(Bucket=bucket_name, Key=base_name)[
                "ContentLength"
            ]
            return int(size)
        except Exception as e:
            raise StorageError("S3", e)

    elif storage_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(tray_name)

        try:
            size, mtime = ioctx.stat(base_name)
            return size
        except Exception as e:
            raise StorageError("CEPH", e)

    elif storage_type == StorageType.FILE:
        try:
            file_stats = os.stat(path)
            return file_stats.st_size
        except Exception as e:
            raise StorageError("FILE", e)

    elif storage_type == StorageType.HTTP or storage_type == StorageType.HTTPS:
        try:
            # Le stream=True permet de ne télécharger que le header initialement
            reponse = requests.get(storage_type.value + path, stream=True).headers["content-length"]
            return reponse
        except Exception as e:
            raise StorageError(storage_type.name, e)

    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to get size")


def exists(path: str) -> bool:
    """Do the file or object exist ?

    Args:
        path (str): path of file/object to test

    Raises:
        MissingEnvironmentError: Missing object storage informations
        StorageError: Storage read issue

    Returns:
        bool: file/object existing status
    """

    storage_type, path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        try:
            s3_client["client"].head_object(Bucket=bucket_name, Key=base_name)
            return True
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise StorageError("S3", e)

    elif storage_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(tray_name)

        try:
            ioctx.stat(base_name)
            return True
        except rados.ObjectNotFound:
            return False
        except Exception as e:
            raise StorageError("CEPH", e)

    elif storage_type == StorageType.FILE:
        return os.path.exists(path)

    elif storage_type == StorageType.HTTP or storage_type == StorageType.HTTPS:
        try:
            response = requests.get(storage_type.value + path, stream=True)
            if response.status_code == 200:
                return True
            else:
                return False
        except Exception as e:
            raise StorageError(storage_type.name, e)

    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to test if exists")


def remove(path: str) -> None:
    """Remove the file/object

    Args:
        path (str): path of file/object to remove

    Raises:
        MissingEnvironmentError: Missing object storage informations
        StorageError: Storage removal issue
    """
    storage_type, path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        try:
            s3_client["client"].delete_object(Bucket=bucket_name, Key=base_name)
        except Exception as e:
            raise StorageError("S3", e)

    elif storage_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(tray_name)

        try:
            ioctx.remove_object(base_name)
        except rados.ObjectNotFound:
            pass
        except Exception as e:
            raise StorageError("CEPH", e)

    elif storage_type == StorageType.FILE:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass
        except Exception as e:
            raise StorageError("FILE", e)

    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to remove things")


def copy(from_path: str, to_path: str, from_md5: str = None) -> None:
    """Copy a file or object to a file or object place. If MD5 sum is provided, it is compared to sum after the copy.

    Args:
        from_path (str): source file/object path, to copy
        to_path (str): destination file/object path
        from_md5 (str, optional): MD5 sum, re-processed after copy and controlled. Defaults to None.

    Raises:
        StorageError: Unhandled copy or copy issue
        MissingEnvironmentError: Missing object storage informations
    """

    from_type, from_path, from_tray, from_base_name = get_infos_from_path(from_path)
    to_type, to_path, to_tray, to_base_name = get_infos_from_path(to_path)

    # Réalisation de la copie, selon les types de stockage
    if from_type == StorageType.FILE and to_type == StorageType.FILE:
        try:
            if to_tray != "":
                os.makedirs(to_tray, exist_ok=True)

            copyfile(from_path, to_path)

            if from_md5 is not None:
                to_md5 = hash_file(to_path)
                if to_md5 != from_md5:
                    raise StorageError(
                        "FILE",
                        f"Invalid MD5 sum control for copy file {from_path} to {to_path} : {from_md5} != {to_md5}",
                    )

        except Exception as e:
            raise StorageError("FILE", f"Cannot copy file {from_path} to {to_path} : {e}")

    elif from_type == StorageType.S3 and to_type == StorageType.FILE:
        s3_client, from_bucket = __get_s3_client(from_tray)

        try:
            if to_tray != "":
                os.makedirs(to_tray, exist_ok=True)

            s3_client["client"].download_file(from_bucket, from_base_name, to_path)

            if from_md5 is not None:
                to_md5 = hash_file(to_path)
                if to_md5 != from_md5:
                    raise StorageError(
                        "S3 and FILE",
                        f"Invalid MD5 sum control for copy S3 object {from_path} to file {to_path} : {from_md5} != {to_md5}",
                    )

        except Exception as e:
            raise StorageError(
                "S3 and FILE", f"Cannot copy S3 object {from_path} to file {to_path} : {e}"
            )

    elif from_type == StorageType.FILE and to_type == StorageType.S3:
        s3_client, to_bucket = __get_s3_client(to_tray)

        try:
            s3_client["client"].upload_file(from_path, to_bucket, to_base_name)

            if from_md5 is not None:
                to_md5 = (
                    s3_client["client"]
                    .head_object(Bucket=to_bucket, Key=to_base_name)["ETag"]
                    .strip('"')
                )
                if to_md5 != from_md5:
                    raise StorageError(
                        "FILE and S3",
                        f"Invalid MD5 sum control for copy file {from_path} to S3 object {to_path} : {from_md5} != {to_md5}",
                    )
        except Exception as e:
            raise StorageError(
                "FILE and S3", f"Cannot copy file {from_path} to S3 object {to_path} : {e}"
            )

    elif from_type == StorageType.S3 and to_type == StorageType.S3:
        from_s3_client, from_bucket = __get_s3_client(from_tray)
        to_s3_client, to_bucket = __get_s3_client(to_tray)

        try:
            if to_s3_client["host"] == from_s3_client["host"]:
                to_s3_client["client"].copy(
                    {"Bucket": from_bucket, "Key": from_base_name}, to_bucket, to_base_name
                )
            else:
                with tempfile.NamedTemporaryFile("w+b") as f:
                    from_s3_client["client"].download_fileobj(from_bucket, from_base_name, f)
                    to_s3_client["client"].upload_file(f.name, to_bucket, to_base_name)

            if from_md5 is not None:
                to_md5 = (
                    to_s3_client["client"]
                    .head_object(Bucket=to_bucket, Key=to_base_name)["ETag"]
                    .strip('"')
                )
                if to_md5 != from_md5:
                    raise StorageError(
                        "S3",
                        f"Invalid MD5 sum control for copy S3 object {from_path} to {to_path} : {from_md5} != {to_md5}",
                    )

        except Exception as e:
            raise StorageError("S3", f"Cannot copy S3 object {from_path} to {to_path} : {e}")

    elif from_type == StorageType.CEPH and to_type == StorageType.FILE:
        ioctx = __get_ceph_ioctx(from_tray)

        if from_md5 is not None:
            checker = hashlib.md5()

        try:
            if to_tray != "":
                os.makedirs(to_tray, exist_ok=True)
            f = open(to_path, "wb")

            offset = 0
            size = 0

            while True:
                chunk = ioctx.read(from_base_name, 65536, offset)
                size = len(chunk)
                offset += size
                f.write(chunk)

                if from_md5 is not None:
                    checker.update(chunk)

                if size < 65536:
                    break

            f.close()

            if from_md5 is not None and from_md5 != checker.hexdigest():
                raise StorageError(
                    "CEPH and FILE",
                    f"Invalid MD5 sum control for copy CEPH object {from_path} to file {to_path} : {from_md5} != {checker.hexdigest()}",
                )

        except Exception as e:
            raise StorageError(
                "CEPH and FILE", f"Cannot copy CEPH object {from_path} to file {to_path} : {e}"
            )

    elif from_type == StorageType.FILE and to_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(to_tray)

        if from_md5 is not None:
            checker = hashlib.md5()

        try:
            f = open(from_path, "rb")

            offset = 0
            size = 0

            while True:
                chunk = f.read(65536)
                size = len(chunk)
                ioctx.write(to_base_name, chunk, offset)
                offset += size

                if from_md5 is not None:
                    checker.update(chunk)

                if size < 65536:
                    break

            f.close()

            if from_md5 is not None and from_md5 != checker.hexdigest():
                raise StorageError(
                    "FILE and CEPH",
                    f"Invalid MD5 sum control for copy file {from_path} to CEPH object {to_path} : {from_md5} != {checker.hexdigest()}",
                )

        except Exception as e:
            raise StorageError(
                "FILE and CEPH", f"Cannot copy file {from_path} to CEPH object {to_path} : {e}"
            )

    elif from_type == StorageType.CEPH and to_type == StorageType.CEPH:
        from_ioctx = __get_ceph_ioctx(from_tray)
        to_ioctx = __get_ceph_ioctx(to_tray)

        if from_md5 is not None:
            checker = hashlib.md5()

        try:
            offset = 0
            size = 0

            while True:
                chunk = from_ioctx.read(from_base_name, 65536, offset)
                size = len(chunk)
                to_ioctx.write(to_base_name, chunk, offset)
                offset += size

                if from_md5 is not None:
                    checker.update(chunk)

                if size < 65536:
                    break

            if from_md5 is not None and from_md5 != checker.hexdigest():
                raise StorageError(
                    "FILE and CEPH",
                    f"Invalid MD5 sum control for copy CEPH object {from_path} to {to_path} : {from_md5} != {checker.hexdigest()}",
                )

        except Exception as e:
            raise StorageError("CEPH", f"Cannot copy CEPH object {from_path} to {to_path} : {e}")

    elif from_type == StorageType.CEPH and to_type == StorageType.S3:
        from_ioctx = __get_ceph_ioctx(from_tray)

        s3_client, to_bucket = __get_s3_client(to_tray)

        if from_md5 is not None:
            checker = hashlib.md5()

        try:
            offset = 0
            size = 0

            with tempfile.NamedTemporaryFile("w+b", delete=False) as f:
                name_tmp = f.name
                while True:
                    chunk = from_ioctx.read(from_base_name, 65536, offset)
                    size = len(chunk)
                    offset += size
                    f.write(chunk)

                    if from_md5 is not None:
                        checker.update(chunk)

                    if size < 65536:
                        break

            s3_client["client"].upload_file(name_tmp, to_bucket, to_base_name)

            os.remove(name_tmp)

            if from_md5 is not None and from_md5 != checker.hexdigest():
                raise StorageError(
                    "CEPH and S3",
                    f"Invalid MD5 sum control for copy CEPH object {from_path} to S3 object {to_path} : {from_md5} != {checker.hexdigest()}",
                )

        except Exception as e:
            raise StorageError(
                "CEPH and S3", f"Cannot copy CEPH object {from_path} to S3 object {to_path} : {e}"
            )

    elif (
        from_type == StorageType.HTTP or from_type == StorageType.HTTPS
    ) and to_type == StorageType.FILE:
        try:
            response = requests.get(from_type.value + from_path, stream=True)
            with open(to_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=65536):
                    if chunk:
                        f.write(chunk)

        except Exception as e:
            raise StorageError(
                "HTTP(S) and FILE",
                f"Cannot copy HTTP(S) object {from_path} to FILE object {to_path} : {e}",
            )

    elif (
        from_type == StorageType.HTTP or from_type == StorageType.HTTPS
    ) and to_type == StorageType.CEPH:
        to_ioctx = __get_ceph_ioctx(to_tray)

        try:
            response = requests.get(from_type.value + from_path, stream=True)
            offset = 0
            for chunk in response.iter_content(chunk_size=65536):
                if chunk:
                    size = len(chunk)
                    to_ioctx.write(to_base_name, chunk, offset)
                    offset += size

        except Exception as e:
            raise StorageError(
                "HTTP(S) and CEPH",
                f"Cannot copy HTTP(S) object {from_path} to CEPH object {to_path} : {e}",
            )

    elif (
        from_type == StorageType.HTTP or from_type == StorageType.HTTPS
    ) and to_type == StorageType.S3:
        to_s3_client, to_bucket = __get_s3_client(to_tray)

        try:
            response = requests.get(from_type.value + from_path, stream=True)
            with tempfile.NamedTemporaryFile("w+b", delete=False) as f:
                name_fich = f.name
                for chunk in response.iter_content(chunk_size=65536):
                    if chunk:
                        f.write(chunk)

            to_s3_client["client"].upload_file(name_fich, to_tray, to_base_name)

            os.remove(name_fich)

        except Exception as e:
            raise StorageError(
                "HTTP(S) and S3",
                f"Cannot copy HTTP(S) object {from_path} to S3 object {to_path} : {e}",
            )

    else:
        raise StorageError(
            f"{from_type.name} and {to_type.name}",
            f"Cannot copy from {from_type.name} to {to_type.name}",
        )


def link(target_path: str, link_path: str, hard: bool = False) -> None:
    """Create a symbolic link

    Args:
        target_path (str): file/object to link
        link_path (str): link to create
        hard (bool, optional): hard link rather than symbolic. Only for FILE storage. Defaults to False.

    Raises:
        StorageError: Unhandled link or link issue
        MissingEnvironmentError: Missing object storage informations
    """

    target_type, target_path, target_tray, target_base_name = get_infos_from_path(target_path)
    link_type, link_path, link_tray, link_base_name = get_infos_from_path(link_path)

    if target_type != link_type:
        raise StorageError(
            f"{target_type.name} and {link_type.name}",
            "Cannot make link between two different storage types",
        )

    if hard and target_type != StorageType.FILE:
        raise StorageError(target_type.name, "Hard link is available only for FILE storage")

    # Réalisation du lien, selon les types de stockage
    if target_type == StorageType.S3:
        target_s3_client, target_bucket = __get_s3_client(target_tray)
        link_s3_client, link_bucket = __get_s3_client(link_tray)

        if target_s3_client["host"] != link_s3_client["host"]:
            raise StorageError(
                "S3",
                f"Cannot make link {link_path} -> {target_path} : link works only on the same S3 cluster",
            )

        try:
            target_s3_client["client"].put_object(
                Body=f"{__OBJECT_SYMLINK_SIGNATURE}{target_bucket}/{target_base_name}".encode(),
                Bucket=link_bucket,
                Key=link_base_name,
            )
        except Exception as e:
            raise StorageError("S3", e)

    elif target_type == StorageType.CEPH:
        ioctx = __get_ceph_ioctx(link_tray)

        try:
            ioctx.write_full(link_base_name, f"{__OBJECT_SYMLINK_SIGNATURE}{target_path}".encode())
        except Exception as e:
            raise StorageError("CEPH", e)

    elif target_type == StorageType.FILE:
        try:
            if hard:
                os.link(target_path, link_path)
            else:
                os.symlink(target_path, link_path)
        except Exception as e:
            raise StorageError("FILE", e)

    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to make link")


def get_osgeo_path(path: str) -> str:
    """Return GDAL/OGR Open compliant path and configure storage access

    For a S3 input path, endpoint, access and secret keys are set and path is built with "/vsis3" root.

    For a FILE input path, only storage prefix is removed

    Args:
        path (str): Source path

    Raises:
        NotImplementedError: Storage type not handled

    Returns:
        str: GDAL/OGR Open compliant path
    """

    storage_type, unprefixed_path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        gdal.SetConfigOption("AWS_SECRET_ACCESS_KEY", s3_client["secret_key"])
        gdal.SetConfigOption("AWS_ACCESS_KEY_ID", s3_client["key"])
        gdal.SetConfigOption("AWS_S3_ENDPOINT", s3_client["host"])
        gdal.SetConfigOption("AWS_VIRTUAL_HOSTING", "FALSE")

        return f"/vsis3/{bucket_name}/{base_name}"

    elif storage_type == StorageType.FILE:
        return unprefixed_path

    else:
        raise NotImplementedError(f"Cannot get a GDAL/OGR compliant path from {path}")


def size_path(path: str) -> int:
    """Return the size of the path given (or, for the CEPH, the sum of the size of each object of the .list)

    Args:
        path (str): Source path

    Raises:
        StorageError: Unhandled link or link issue
        MissingEnvironmentError: Missing object storage informations

    Returns:
        int: size of the path
    """
    storage_type, unprefixed_path, tray_name, base_name = get_infos_from_path(path)

    if storage_type == StorageType.FILE:
        try:
            total = 0
            with os.scandir(unprefixed_path) as it:
                for entry in it:
                    if entry.is_file():
                        total += entry.stat().st_size
                    elif entry.is_dir():
                        total += size_path(entry.path)

        except Exception as e:
            raise StorageError("FILE", e)

    elif storage_type == StorageType.S3:
        s3_client, bucket_name = __get_s3_client(tray_name)

        try:
            paginator = s3_client["client"].get_paginator("list_objects_v2")
            pages = paginator.paginate(
                Bucket=bucket_name,
                Prefix=base_name + "/",
                PaginationConfig={
                    "PageSize": 10000,
                },
            )
            total = 0
            for page in pages:
                for key in page["Contents"]:
                    total += key["Size"]

        except Exception as e:
            raise StorageError("S3", e)

    elif storage_type == StorageType.CEPH:
        raise NotImplementedError
    else:
        raise StorageError("UNKNOWN", "Unhandled storage type to calculate size")

    return total



import os
from unittest import mock
from unittest.mock import *

import pytest

from rok4.enums import SlabType, StorageType
from rok4.exceptions import *
from rok4.pyramid import *
from rok4.utils import *


@mock.patch("rok4.pyramid.get_data_str", side_effect=StorageError("FILE", "Not found"))
def test_wrong_file(mocked_get_data_str):
    with pytest.raises(StorageError):
        Pyramid.from_descriptor("file:///pyramid.json")


@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"format": "TIFF_PBF_MVT","levels":[{"id": "100","tables":',
)
def test_bad_json(mocked_get_data_str):
    with pytest.raises(FormatError) as exc:
        Pyramid.from_descriptor("file:///pyramid.json")

    assert (
        str(exc.value)
        == "Expected format JSON to read 'file:///pyramid.json' : Expecting value: line 1 column 59 (char 58)"
    )
    mocked_get_data_str.assert_called_once_with("file:///pyramid.json")


@mock.patch("rok4.pyramid.get_data_str", return_value='{"format": "TIFF_PBF_MVT","levels":[]}')
def test_missing_tms(mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        Pyramid.from_descriptor("file:///pyramid.json")

    assert str(exc.value) == "Missing attribute 'tile_matrix_set' in 'file:///pyramid.json'"
    mocked_get_data_str.assert_called_once_with("file:///pyramid.json")


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"format": "TIFF_PBF_MVT","levels":[{}], "tile_matrix_set": "PM"}',
)
@mock.patch("rok4.pyramid.TileMatrixSet", side_effect=StorageError("FILE", "TMS not found"))
def test_wrong_tms(mocked_tms_constructor, mocked_get_data_str):
    with pytest.raises(StorageError) as exc:
        Pyramid.from_descriptor("file:///pyramid.json")

    assert str(exc.value) == "Issue occured using a FILE storage : TMS not found"
    mocked_tms_constructor.assert_called_once_with("PM")
    mocked_get_data_str.assert_called_once_with("file:///pyramid.json")


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"format": "TIFF_JPG_UINT8","levels":[{"tiles_per_height":16,"tile_limits":{"min_col":0,"max_row":15,"max_col":15,"min_row":0},"storage":{"image_directory":"SCAN1000/DATA/0","path_depth":2,"type":"FILE"},"tiles_per_width":16,"id":"0"}], "tile_matrix_set": "PM"}',
)
@mock.patch("rok4.pyramid.TileMatrixSet")
def test_raster_missing_raster_specifications(mocked_tms_class, mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        Pyramid.from_descriptor("file:///pyramid.json")

    assert str(exc.value) == "Missing attribute 'raster_specifications' in 'file:///pyramid.json'"
    mocked_get_data_str.assert_called_once_with("file:///pyramid.json")


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"raster_specifications":{"channels":3,"nodata":"255,0,0","photometric":"rgb","interpolation":"bicubic"}, "format": "TIFF_JPG_UINT8","levels":[{"tiles_per_height":16,"tile_limits":{"min_col":0,"max_row":15,"max_col":15,"min_row":0},"storage":{"image_directory":"SCAN1000/DATA/0","path_depth":2,"type":"FILE"},"tiles_per_width":16,"id":"unknown"}], "tile_matrix_set": "PM"}',
)
@mock.patch("rok4.pyramid.TileMatrixSet")
def test_wrong_level(mocked_tms_class, mocked_get_data_str):
    tms_instance = MagicMock()
    tms_instance.get_level.return_value = None
    tms_instance.name = "PM"
    mocked_tms_class.return_value = tms_instance

    with pytest.raises(Exception) as exc:
        Pyramid.from_descriptor("file:///pyramid.json")

    mocked_tms_class.assert_called_once_with("PM")
    mocked_get_data_str.assert_called_once_with("file:///pyramid.json")
    tms_instance.get_level.assert_called_once_with("unknown")
    assert (
        str(exc.value)
        == "Pyramid file:///pyramid.json owns a level with the ID 'unknown', not defined in the TMS 'PM'"
    )


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"format": "TIFF_PBF_MVT","levels":[{"tiles_per_height":16,"tile_limits":{"min_col":0,"max_row":15,"max_col":15,"min_row":0},"storage":{"image_directory":"SCAN1000/DATA/0","path_depth":2,"type":"FILE"},"tiles_per_width":16,"id":"0"}], "tile_matrix_set": "PM"}',
)
@mock.patch("rok4.pyramid.TileMatrixSet", autospec=True)
def test_vector_missing_tables(mocked_tms_class, mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        Pyramid.from_descriptor("file:///pyramid.json")

    assert str(exc.value) == "Missing attribute levels[].'tables' in 'file:///pyramid.json'"
    mocked_get_data_str.assert_called_once_with("file:///pyramid.json")


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"raster_specifications":{"channels":3,"nodata":"255,0,0","photometric":"rgb","interpolation":"bicubic"}, "format": "TIFF_JPG_UINT8","levels":[{"tiles_per_height":16,"tile_limits":{"min_col":0,"max_row":15,"max_col":15,"min_row":0},"storage":{"image_prefix":"SCAN1000/DATA_0","pool_name":"pool1","type":"CEPH"},"tiles_per_width":16,"id":"0"}], "tile_matrix_set": "PM"}',
)
@mock.patch("rok4.pyramid.TileMatrixSet")
@mock.patch("rok4.pyramid.put_data_str", return_value=None)
def test_raster_ok(mocked_put_data_str, mocked_tms_class, mocked_get_data_str):
    tms_instance = MagicMock()
    tms_instance.name = "PM"
    tms_instance.srs = "EPSG:3857"
    tms_instance.sr = srs_to_spatialreference("EPSG:3857")

    tm_instance = MagicMock()
    tm_instance.id = "0"
    tm_instance.resolution = 1
    tm_instance.point_to_indices.return_value = (0, 0, 128, 157)

    tms_instance.get_level.return_value = tm_instance

    mocked_tms_class.return_value = tms_instance

    try:
        pyramid = Pyramid.from_descriptor("ceph://pool1/sub/pyramid.json")
        assert pyramid.get_level("0") is not None
        assert pyramid.get_level("4") is None
        assert pyramid.name == "sub/pyramid"
        assert pyramid.storage_type == StorageType.CEPH
        assert pyramid.storage_root == "pool1"
        mocked_get_data_str.assert_called_once_with("ceph://pool1/sub/pyramid.json")

        clone = Pyramid.from_other(pyramid, "titi", {"type": "FILE", "root": "/data/ign"})
        assert clone.name == "titi"
        assert clone.storage_type == StorageType.FILE
        assert clone.storage_root == "/data/ign"
        assert clone.tile_extension == "jpg"
        assert clone.get_level("0") is not None
        assert clone.get_level("4") is None
        assert clone.get_infos_from_slab_path("/data/ign/titi/IMAGE/12/00/4A/F7.tif") == (
            SlabType.DATA,
            "12",
            159,
            367,
        )
        assert clone.get_tile_indices(102458, 6548125, srs="EPSG:3857") == ("0", 0, 0, 128, 157)
        assert clone.get_tile_indices(43, 2, srs="EPSG:4326") == ("0", 0, 0, 128, 157)

        assert len(clone.get_levels()) == 1

        clone.write_descriptor()
        mocked_put_data_str.assert_called_once_with(
            '{"tile_matrix_set": "PM", "format": "TIFF_JPG_UINT8", "levels": [{"id": "0", "tiles_per_width": 16, "tiles_per_height": 16, "tile_limits": {"min_col": 0, "max_row": 15, "max_col": 15, "min_row": 0}, "storage": {"type": "FILE", "image_directory": "titi/DATA/0", "path_depth": 2}}], "raster_specifications": {"channels": 3, "nodata": "255,0,0", "photometric": "rgb", "interpolation": "bicubic"}}',
            "file:///data/ign/titi.json",
        )
    except Exception as exc:
        assert False, f"Pyramid creation raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch(
    "rok4.pyramid.get_data_str",
    return_value='{"format": "TIFF_PBF_MVT","levels":[{"tiles_per_height":16,"tile_limits":{"min_col":0,"max_row":15,"max_col":15,"min_row":0},"storage":{"image_directory":"SCAN1000/DATA/0","path_depth":2,"type":"FILE"},"tiles_per_width":16,"id":"0","tables":[{"name":"table","geometry":"POINT","attributes":[{"type":"bigint","name":"fid","count":1531}]}]}], "tile_matrix_set": "PM"}',
)
@mock.patch("rok4.pyramid.TileMatrixSet")
def test_vector_ok(mocked_tms_class, mocked_get_data_str):
    try:
        pyramid = Pyramid.from_descriptor("file:///pyramid.json")
        assert pyramid.get_level("0") is not None
        assert pyramid.get_level("4") is None
        assert pyramid.name == "pyramid"
        assert pyramid.storage_depth == 2
        assert pyramid.storage_type == StorageType.FILE
        mocked_get_data_str.assert_called_once_with("file:///pyramid.json")

        clone = Pyramid.from_other(pyramid, "toto", {"type": "S3", "root": "bucket"})
        assert clone.name == "toto"
        assert clone.tile_extension == "pbf"
        assert clone.storage_type == StorageType.S3
        assert clone.get_level("0") is not None
        assert clone.get_level("4") is None
    except Exception as exc:
        assert False, f"Pyramid creation raises an exception: {exc}"

    with pytest.raises(Exception):
        pyramid.get_tile_data_raster("12", 5, 6)


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("rok4.pyramid.TileMatrixSet")
def test_tile_read_raster(mocked_tms_class):
    tms_instance = MagicMock()
    tms_instance.name = "UTM20W84MART_1M_MNT"
    tms_instance.srs = "IGNF:UTM20W84MART"

    tm_instance = MagicMock()
    tm_instance.id = "8"
    tm_instance.tile_size = (256, 256)

    tms_instance.get_level.return_value = tm_instance

    mocked_tms_class.return_value = tms_instance

    try:
        pyramid = Pyramid.from_descriptor("file://tests/fixtures/TIFF_ZIP_FLOAT32.json")
        data = pyramid.get_tile_data_raster("8", 2748, 40537)

        assert data.shape == (256, 256, 1)
        assert data[128][128][0] == 447.25
    except Exception as exc:
        assert False, f"Pyramid raster tile read raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("rok4.pyramid.TileMatrixSet")
def test_tile_read_vector(mocked_tms_class):
    tms_instance = MagicMock()
    tms_instance.name = "PM"
    tms_instance.srs = "EPSG:3857"

    tm_instance = MagicMock()
    tm_instance.id = "4"
    tm_instance.tile_size = (256, 256)

    tms_instance.get_level.return_value = tm_instance

    mocked_tms_class.return_value = tms_instance

    try:
        pyramid = Pyramid.from_descriptor("file://tests/fixtures/TIFF_PBF_MVT.json")

        data = pyramid.get_tile_data_vector("4", 5, 5)
        assert data is None

        data = pyramid.get_tile_data_vector("4", 8, 5)
        assert type(data) is dict
        assert "ecoregions_3857" in data
    except Exception as exc:
        assert False, f"Pyramid vector tile read raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("rok4.pyramid.TileMatrixSet")
def test_list_read(mocked_tms_class):
    tms_instance = MagicMock()
    tms_instance.name = "PM"
    tms_instance.srs = "EPSG:3857"

    tm_instance = MagicMock()
    tm_instance.id = "4"
    tm_instance.tile_size = (256, 256)

    tms_instance.get_level.return_value = tm_instance

    mocked_tms_class.return_value = tms_instance

    try:
        pyramid = Pyramid.from_descriptor("file://tests/fixtures/TIFF_PBF_MVT.json")
        pyramid.load_list()
        pyramid.load_list()  # on passe par la détection d'une liste déjà chrargée ainsi
        for (slab_type, level, column, row), infos in pyramid.list_generator():
            assert slab_type == SlabType.DATA
            assert level == "4"
            assert column == 2
            assert row == 1
            assert infos == {
                "link": False,
                "md5": None,
                "root": "TIFF_PBF_MVT",
                "slab": "DATA/4/00/00/21.tif",
            }

    except Exception as exc:
        assert False, f"Pyramid vector list read raises an exception: {exc}"


def test_b36_path_decode():
    assert b36_path_decode("3E/42/01.tif") == (
        4032,
        18217,
    )
    assert b36_path_decode("3E/42/01.TIFF") == (
        4032,
        18217,
    )
    assert b36_path_decode("3E/42/01") == (
        4032,
        18217,
    )


def test_b36_path_encode():
    assert b36_path_encode(4032, 18217, 2) == "3E/42/01.tif"
    assert b36_path_encode(14, 18217, 1) == "0E02/E1.tif"



"""Provide classes to use a layer.

The module contains the following classe:

- `Layer` - Descriptor to broadcast pyramids' data
"""

# -- IMPORTS --

# standard library
import json
import os
import re
from json.decoder import JSONDecodeError
from typing import Dict, List, Tuple

# package
from rok4.enums import PyramidType
from rok4.exceptions import FormatError, MissingAttributeError
from rok4.pyramid import Pyramid
from rok4.storage import get_data_str, get_infos_from_path, put_data_str
from rok4.utils import reproject_bbox


class Layer:
    """A data layer, raster or vector

    Attributes:
        __name (str): layer's technical name
        __pyramids (Dict[str, Union[rok4.pyramid.Pyramid,str,str]]): used pyramids
        __format (str): pyramid's list path
        __tms (rok4.tile_matrix_set.TileMatrixSet): Used grid
        __keywords (List[str]): Keywords
        __levels (Dict[str, rok4.pyramid.Level]): Used pyramids' levels
        __best_level (rok4.pyramid.Level): Used pyramids best level
        __resampling (str): Interpolation to use fot resampling
        __bbox (Tuple[float, float, float, float]): data bounding box, TMS coordinates system
        __geobbox (Tuple[float, float, float, float]): data bounding box, EPSG:4326
    """

    @classmethod
    def from_descriptor(cls, descriptor: str) -> "Layer":
        """Create a layer from its descriptor

        Args:
            descriptor (str): layer's descriptor path

        Raises:
            FormatError: Provided path is not a well formed JSON
            MissingAttributeError: Attribute is missing in the content
            StorageError: Storage read issue (layer descriptor)
            MissingEnvironmentError: Missing object storage informations

        Returns:
            Layer: a Layer instance
        """
        try:
            data = json.loads(get_data_str(descriptor))

        except JSONDecodeError as e:
            raise FormatError("JSON", descriptor, e)

        layer = cls()

        storage_type, path, root, base_name = get_infos_from_path(descriptor)
        layer.__name = base_name[:-5]  # on supprime l'extension.json

        try:
            # Attributs communs
            layer.__title = data["title"]
            layer.__abstract = data["abstract"]
            layer.__load_pyramids(data["pyramids"])

            # Paramètres optionnels
            if "keywords" in data:
                for k in data["keywords"]:
                    layer.__keywords.append(k)

            if layer.type == PyramidType.RASTER:
                if "resampling" in data:
                    layer.__resampling = data["resampling"]

                if "styles" in data:
                    layer.__styles = data["styles"]
                else:
                    layer.__styles = ["normal"]

            # Les bbox, native et géographique
            if "bbox" in data:
                layer.__geobbox = (
                    data["bbox"]["south"],
                    data["bbox"]["west"],
                    data["bbox"]["north"],
                    data["bbox"]["east"],
                )
                layer.__bbox = reproject_bbox(layer.__geobbox, "EPSG:4326", layer.__tms.srs, 5)
                # On force l'emprise de la couche, on recalcule donc les tuiles limites correspondantes pour chaque niveau
                for level in layer.__levels.values():
                    level.set_limits_from_bbox(layer.__bbox)
            else:
                layer.__bbox = layer.__best_level.bbox
                layer.__geobbox = reproject_bbox(layer.__bbox, layer.__tms.srs, "EPSG:4326", 5)

        except KeyError as e:
            raise MissingAttributeError(descriptor, e)

        return layer

    @classmethod
    def from_parameters(cls, pyramids: List[Dict[str, str]], name: str, **kwargs) -> "Layer":
        """Create a default layer from parameters

        Args:
            pyramids (List[Dict[str, str]]): pyramids to use and extrem levels, bottom and top
            name (str): layer's technical name
            **title (str): Layer's title (will be equal to name if not provided)
            **abstract (str): Layer's abstract (will be equal to name if not provided)
            **styles (List[str]): Styles identifier to authorized for the layer
            **resampling (str): Interpolation to use for resampling

        Raises:
            Exception: name contains forbidden characters or used pyramids do not shared same parameters (format, tms...)

        Returns:
            Layer: a Layer instance
        """

        layer = cls()

        # Informations obligatoires
        if not re.match("^[A-Za-z0-9_-]*$", name):
            raise Exception(
                f"Layer's name have to contain only letters, number, hyphen and underscore, to be URL and storage compliant ({name})"
            )

        layer.__name = name
        layer.__load_pyramids(pyramids)

        # Les bbox, native et géographique
        layer.__bbox = layer.__best_level.bbox
        layer.__geobbox = reproject_bbox(layer.__bbox, layer.__tms.srs, "EPSG:4326", 5)

        # Informations calculées
        layer.__keywords.append(layer.type.name)
        layer.__keywords.append(layer.__name)

        # Informations optionnelles
        if "title" in kwargs and kwargs["title"] is not None:
            layer.__title = kwargs["title"]
        else:
            layer.__title = name

        if "abstract" in kwargs and kwargs["abstract"] is not None:
            layer.__abstract = kwargs["abstract"]
        else:
            layer.__abstract = name

        if layer.type == PyramidType.RASTER:
            if "styles" in kwargs and kwargs["styles"] is not None and len(kwargs["styles"]) > 0:
                layer.__styles = kwargs["styles"]
            else:
                layer.__styles = ["normal"]

            if "resampling" in kwargs and kwargs["resampling"] is not None:
                layer.__resampling = kwargs["resampling"]

        return layer

    def __init__(self) -> None:
        self.__format = None
        self.__tms = None
        self.__best_level = None
        self.__levels = {}
        self.__keywords = []
        self.__pyramids = []

    def __load_pyramids(self, pyramids: List[Dict[str, str]]) -> None:
        """Load and check pyramids

        Args:
            pyramids (List[Dict[str, str]]): List of descriptors' paths and optionnaly top and bottom levels

        Raises:
            Exception: Pyramids' do not all own the same format
            Exception: Pyramids' do not all own the same TMS
            Exception: Pyramids' do not all own the same channels number
            Exception: Overlapping in usage pyramids' levels
        """

        # Toutes les pyramides doivent avoir les même caractéristiques
        channels = None
        for p in pyramids:
            pyramid = Pyramid.from_descriptor(p["path"])
            bottom_level = p.get("bottom_level", None)
            top_level = p.get("top_level", None)

            if bottom_level is None:
                bottom_level = pyramid.bottom_level.id

            if top_level is None:
                top_level = pyramid.top_level.id

            if self.__format is not None and self.__format != pyramid.format:
                raise Exception(
                    f"Used pyramids have to own the same format : {self.__format} != {pyramid.format}"
                )
            else:
                self.__format = pyramid.format

            if self.__tms is not None and self.__tms.id != pyramid.tms.id:
                raise Exception(
                    f"Used pyramids have to use the same TMS : {self.__tms.id} != {pyramid.tms.id}"
                )
            else:
                self.__tms = pyramid.tms

            if self.type == PyramidType.RASTER:
                if channels is not None and channels != pyramid.raster_specifications["channels"]:
                    raise Exception(
                        f"Used RASTER pyramids have to own the same number of channels : {channels} != {pyramid.raster_specifications['channels']}"
                    )
                else:
                    channels = pyramid.raster_specifications["channels"]
                self.__resampling = pyramid.raster_specifications["interpolation"]

            levels = pyramid.get_levels(bottom_level, top_level)
            for level in levels:
                if level.id in self.__levels:
                    raise Exception(f"Level {level.id} is present in two used pyramids")
                self.__levels[level.id] = level

            self.__pyramids.append(
                {"pyramid": pyramid, "bottom_level": bottom_level, "top_level": top_level}
            )

        self.__best_level = sorted(self.__levels.values(), key=lambda level: level.resolution)[0]

    def __str__(self) -> str:
        return f"{self.type.name} layer '{self.__name}'"

    @property
    def serializable(self) -> Dict:
        """Get the dict version of the layer object, descriptor compliant

        Returns:
            Dict: descriptor structured object description
        """
        serialization = {
            "title": self.__title,
            "abstract": self.__abstract,
            "keywords": self.__keywords,
            "wmts": {"authorized": True},
            "tms": {"authorized": True},
            "bbox": {
                "south": self.__geobbox[0],
                "west": self.__geobbox[1],
                "north": self.__geobbox[2],
                "east": self.__geobbox[3],
            },
            "pyramids": [],
        }

        for p in self.__pyramids:
            serialization["pyramids"].append(
                {
                    "bottom_level": p["bottom_level"],
                    "top_level": p["top_level"],
                    "path": p["pyramid"].descriptor,
                }
            )

        if self.type == PyramidType.RASTER:
            serialization["wms"] = {
                "authorized": True,
                "crs": ["CRS:84", "IGNF:WGS84G", "EPSG:3857", "EPSG:4258", "EPSG:4326"],
            }

            if self.__tms.srs.upper() not in serialization["wms"]["crs"]:
                serialization["wms"]["crs"].append(self.__tms.srs.upper())

            serialization["styles"] = self.__styles
            serialization["resampling"] = self.__resampling

        return serialization

    def write_descriptor(self, directory: str = None) -> None:
        """Print layer's descriptor as JSON

        Args:
            directory (str, optional): Directory (file or object) where to print the layer's descriptor, called <layer's name>.json. Defaults to None, JSON is printed to standard output.
        """
        content = json.dumps(self.serializable)

        if directory is None:
            print(content)
        else:
            put_data_str(content, os.path.join(directory, f"{self.__name}.json"))

    @property
    def type(self) -> PyramidType:
        if self.__format == "TIFF_PBF_MVT":
            return PyramidType.VECTOR
        else:
            return PyramidType.RASTER

    @property
    def bbox(self) -> Tuple[float, float, float, float]:
        return self.__bbox

    @property
    def geobbox(self) -> Tuple[float, float, float, float]:
        return self.__geobbox



"""Provide classes to use a tile matrix set.

The module contains the following classes:

- `TileMatrixSet` - Multi level grid
- `TileMatrix` - A tile matrix set level

Loading a tile matrix set requires environment variables :
- ROK4_TMS_DIRECTORY
"""

# -- IMPORTS --

# standard library
import json
import os
from json.decoder import JSONDecodeError
from typing import Dict, List, Tuple

# package
from rok4.exceptions import FormatError, MissingAttributeError, MissingEnvironmentError
from rok4.storage import get_data_str
from rok4.utils import srs_to_spatialreference

# -- GLOBALS --


class TileMatrix:
    """A tile matrix is a tile matrix set's level.

    Attributes:
        id (str): TM identifiant (no underscore).
        tms (TileMatrixSet): TMS to whom it belong
        resolution (float): Ground size of a pixel, using unity of the TMS's coordinate system.
        origin (Tuple[float, float]): X,Y coordinates of the upper left corner for the level, the grid's origin.
        tile_size (Tuple[int, int]): Pixel width and height of a tile.
        matrix_size (Tuple[int, int]): Number of tile in the level, widthwise and heightwise.
    """

    def __init__(self, level: Dict, tms: "TileMatrixSet") -> None:
        """Constructor method

        Args:
            level: Level attributes, according to JSON structure
            tms: TMS object containing the level to create

        Raises:
            MissingAttributeError: Attribute is missing in the content
        """

        self.tms = tms
        try:
            self.id = level["id"]
            if self.id.find("_") != -1:
                raise Exception(
                    f"TMS {tms.path} owns a level whom id contains an underscore ({self.id})"
                )
            self.resolution = level["cellSize"]
            self.origin = (
                level["pointOfOrigin"][0],
                level["pointOfOrigin"][1],
            )
            self.tile_size = (
                level["tileWidth"],
                level["tileHeight"],
            )
            self.matrix_size = (
                level["matrixWidth"],
                level["matrixHeight"],
            )
            self.__latlon = (
                self.tms.sr.EPSGTreatsAsLatLong() or self.tms.sr.EPSGTreatsAsNorthingEasting()
            )
        except KeyError as e:
            raise MissingAttributeError(tms.path, f"tileMatrices[].{e}")

    def x_to_column(self, x: float) -> int:
        """Convert west-east coordinate to tile's column

        Args:
            x (float): west-east coordinate (TMS coordinates system)

        Returns:
            int: tile's column
        """
        return int((x - self.origin[0]) / (self.resolution * self.tile_size[0]))

    def y_to_row(self, y: float) -> int:
        """Convert north-south coordinate to tile's row

        Args:
            y (float): north-south coordinate (TMS coordinates system)

        Returns:
            int: tile's row
        """
        return int((self.origin[1] - y) / (self.resolution * self.tile_size[1]))

    def tile_to_bbox(self, tile_col: int, tile_row: int) -> Tuple[float, float, float, float]:
        """Get tile terrain extent (xmin, ymin, xmax, ymax), in TMS coordinates system

        TMS spatial reference is Lat / Lon case is handled.

        Args:
            tile_col (int): column indice
            tile_row (int): row indice

        Returns:
            Tuple[float, float, float, float]: terrain extent (xmin, ymin, xmax, ymax)
        """
        if self.__latlon:
            return (
                self.origin[1] - self.resolution * (tile_row + 1) * self.tile_size[1],
                self.origin[0] + self.resolution * tile_col * self.tile_size[0],
                self.origin[1] - self.resolution * tile_row * self.tile_size[1],
                self.origin[0] + self.resolution * (tile_col + 1) * self.tile_size[0],
            )
        else:
            return (
                self.origin[0] + self.resolution * tile_col * self.tile_size[0],
                self.origin[1] - self.resolution * (tile_row + 1) * self.tile_size[1],
                self.origin[0] + self.resolution * (tile_col + 1) * self.tile_size[0],
                self.origin[1] - self.resolution * tile_row * self.tile_size[1],
            )

    def bbox_to_tiles(self, bbox: Tuple[float, float, float, float]) -> Tuple[int, int, int, int]:
        """Get extrems tile columns and rows corresponding to provided bounding box

        TMS spatial reference is Lat / Lon case is handled.

        Args:
            bbox (Tuple[float, float, float, float]): bounding box (xmin, ymin, xmax, ymax), in TMS coordinates system

        Returns:
            Tuple[int, int, int, int]: extrem tiles (col_min, row_min, col_max, row_max)
        """

        if self.__latlon:
            return (
                self.x_to_column(bbox[1]),
                self.y_to_row(bbox[2]),
                self.x_to_column(bbox[3]),
                self.y_to_row(bbox[0]),
            )
        else:
            return (
                self.x_to_column(bbox[0]),
                self.y_to_row(bbox[3]),
                self.x_to_column(bbox[2]),
                self.y_to_row(bbox[1]),
            )

    def point_to_indices(self, x: float, y: float) -> Tuple[int, int, int, int]:
        """Get pyramid's tile and pixel indices from point's coordinates

        TMS spatial reference with Lat / Lon order is handled.

        Args:
            x (float): point's x
            y (float): point's y

        Returns:
            Tuple[int, int, int, int]: tile's column, tile's row, pixel's (in the tile) column, pixel's row
        """

        if self.__latlon:
            absolute_pixel_column = int((y - self.origin[0]) / self.resolution)
            absolute_pixel_row = int((self.origin[1] - x) / self.resolution)
        else:
            absolute_pixel_column = int((x - self.origin[0]) / self.resolution)
            absolute_pixel_row = int((self.origin[1] - y) / self.resolution)

        return (
            absolute_pixel_column // self.tile_size[0],
            absolute_pixel_row // self.tile_size[1],
            absolute_pixel_column % self.tile_size[0],
            absolute_pixel_row % self.tile_size[1],
        )


class TileMatrixSet:
    """A tile matrix set is multi levels grid definition

    Attributes:
        name (str): TMS's name
        path (str): TMS origin path (JSON)
        id (str): TMS identifier
        srs (str): TMS coordinates system
        sr (osgeo.osr.SpatialReference): TMS OSR spatial reference
        levels (Dict[str, TileMatrix]): TMS levels
    """

    def __init__(self, name: str) -> None:
        """Constructor method

        Args:
            name: TMS's name

        Raises:
            MissingEnvironmentError: Missing object storage informations
            Exception: No level in the TMS, CRS not recognized by OSR
            StorageError: Storage read issue
            FormatError: Provided path is not a well formed JSON
            MissingAttributeError: Attribute is missing in the content
        """

        self.name = name

        try:
            self.path = os.path.join(os.environ["ROK4_TMS_DIRECTORY"], f"{self.name}.json")
        except KeyError as e:
            raise MissingEnvironmentError(e)

        try:
            data = json.loads(get_data_str(self.path))

            self.id = data["id"]
            self.srs = data["crs"]
            self.sr = srs_to_spatialreference(self.srs)
            self.levels = {}
            for level in data["tileMatrices"]:
                lev = TileMatrix(level, self)
                self.levels[lev.id] = lev

            if len(self.levels.keys()) == 0:
                raise Exception(f"TMS '{self.path}' has no level")

            if data["orderedAxes"] != ["X", "Y"] and data["orderedAxes"] != ["Lon", "Lat"]:
                raise Exception(
                    f"TMS '{self.path}' own invalid axes order : only X/Y or Lon/Lat are handled"
                )

        except JSONDecodeError as e:
            raise FormatError("JSON", self.path, e)

        except KeyError as e:
            raise MissingAttributeError(self.path, e)

        except RuntimeError as e:
            raise Exception(
                f"Wrong attribute 'crs' ('{self.srs}') in '{self.path}', not recognize by OSR. Trace : {e}"
            )

    def get_level(self, level_id: str) -> "TileMatrix":
        """Get one level according to its identifier

        Args:
            level_id: Level identifier

        Returns:
            The corresponding tile matrix, None if not present
        """

        return self.levels.get(level_id, None)

    @property
    def sorted_levels(self) -> List[TileMatrix]:
        return sorted(self.levels.values(), key=lambda level: level.resolution)
